# -*- coding: utf-8 -*-
from django.db.models import signals
from django.utils.thread_support import currentThread

_urls_to_purge = {}

def add_purge_url(url):
    _urls_to_purge.setdefault(currentThread(), set()).add(url)

def get_urls():
    return _urls_to_purge.get(currentThread(), set())

def clear_urls():
    try:
        del _urls_to_purge[currentThread()]
    except KeyError:
        pass

def purge_instance_urls(sender, instance, **kwargs):
    """ callback for django post_save / post_delete signal
    request url purge when a model is changed or deleted
    """
    if getattr(instance, 'get_absolute_url', None) is not None:
        add_purge_url(instance.get_absolute_url())

    if getattr(instance, 'get_purge_urls', None) is not None:
        for url in instance.get_purge_urls():
            add_purge_url(url)

def register_model(sender, **kwargs):
    """ register post_save / post_delete callback for models with appropriate
    methods
    """
    if (getattr(sender, 'get_absolute_url', None) is None
        and getattr(sender, 'get_purge_urls', None) is None):
        return

    signals.post_save.connect(purge_instance_urls, sender=sender)
    signals.post_delete.connect(purge_instance_urls, sender=sender)

signals.class_prepared.connect(register_model)



# -*- coding: utf-8 -*-
from django.conf import settings
from django_cachepurge import clear_urls, get_urls
from django_cachepurge.utils import pruneAsync

CACHE_URLS = settings.CACHE_URLS

if isinstance(CACHE_URLS, basestring):
    CACHE_URLS = (CACHE_URLS,)

class CachePurge(object):
    """ Middleware responsible of sending PURGE to upstream cache(s)
    """
    def process_request(self, request):
        clear_urls()

    def process_response(self, request, response):

        for url in get_urls():
            for cache in CACHE_URLS:
                url_in_cache = "%s%s" % (cache, url)

                pruneAsync(url_in_cache)

        return response



# -*- coding: utf-8 -*-
import sys
import new
import unittest
from django.db.models import loading
from django.core.management import call_command
from django.test import TestCase

import django_cachepurge
from django_cachepurge import middleware

# from . import testing models # only python 2.5+
from django_cachepurge.tests import testingmodels

class CachePurgeTestCase(TestCase):

    urls = 'django_cachepurge.tests.urls'

    def _pre_setup(self):
        # register tests models
        # we'll fake an app named 'cachepurge_test_app'
        module = new.module('cachepurge_test_app')
        module.__file__ = __file__ # somewhere django looks at __file__. Feed it.
        module.models = testingmodels
        testingmodels.__name__ = 'cachepurge_test_app.models'
        sys.modules['cachepurge_test_app'] = module

        # register fake app in django and create DB models
        from django.conf import settings
        settings.INSTALLED_APPS += ('cachepurge_test_app',)
        loading.load_app('cachepurge_test_app')
        call_command('syncdb', verbosity=0, interactive=False)

        return super(CachePurgeTestCase, self)._pre_setup()

    def setUp(self):
        super(CachePurgeTestCase, self).setUp()
        django_cachepurge.clear_urls()

        # patch middleware pruneAsync to catch url calls
        self._original_pruneAsync = middleware.pruneAsync
        self._pruned_urls = []
        middleware.pruneAsync = self.addPrunedUrl

    def tearDown(self):
        # restore patched method
        middleware.pruneAsync = self._original_pruneAsync
        super(CachePurgeTestCase, self).tearDown()

    def addPrunedUrl(self, url):
        self._pruned_urls.append(url)

    def clearPrunedUrls(self):
        self._pruned_urls = []

    @property
    def pruned_urls(self):
        return sorted(self._pruned_urls)

    def test_urls_accumulator(self):
        self.assertEquals(django_cachepurge.get_urls(), set())

        # add 1 url
        django_cachepurge.add_purge_url('/some/url')
        self.assertEquals(django_cachepurge.get_urls(), set(('/some/url',)))

        # add url twice
        django_cachepurge.add_purge_url('/some/url')
        self.assertEquals(django_cachepurge.get_urls(), set(('/some/url',)))

        # clear
        django_cachepurge.clear_urls()
        self.assertEquals(django_cachepurge.get_urls(), set())

    def test_signals(self):

        self.assertEquals(django_cachepurge.get_urls(), set())

        # create
        content = testingmodels.UrlModel(title=u'model with url')
        content.save()
        self.assertEquals(sorted(django_cachepurge.get_urls()),
                          ['/content/1', '/some/related/url'])

        # edit
        django_cachepurge.clear_urls()
        content.title = u'edited title'
        self.assertEquals(django_cachepurge.get_urls(), set())
        content.save()
        self.assertEquals(sorted(django_cachepurge.get_urls()),
                          ['/content/1', '/some/related/url'])

        # delete
        django_cachepurge.clear_urls()
        content.delete()
        self.assertEquals(sorted(django_cachepurge.get_urls()),
                          ['/content/1', '/some/related/url'])

    def test_no_url_model_signals(self):
        self.assertEquals(django_cachepurge.get_urls(), set())

        # create
        content = testingmodels.NoUrlModel(title=u'model without url')
        content.save()
        self.assertEquals(django_cachepurge.get_urls(), set())

        content.title = u'edited title'
        content.save()
        self.assertEquals(django_cachepurge.get_urls(), set())

        content.delete()
        self.assertEquals(django_cachepurge.get_urls(), set())

    def test_purge_url_only_model_signals(self):
        self.assertEquals(django_cachepurge.get_urls(), set())
        # create
        content = testingmodels.OnlyPurgeUrls(title=u'model with purge url only')
        content.save()
        self.assertEquals(sorted(django_cachepurge.get_urls()),
                          ['/some/other/url'])

        # edit
        django_cachepurge.clear_urls()
        content.title = u'edited title'
        self.assertEquals(django_cachepurge.get_urls(), set())
        content.save()
        self.assertEquals(sorted(django_cachepurge.get_urls()),
                          ['/some/other/url'])

        # delete
        django_cachepurge.clear_urls()
        content.delete()
        self.assertEquals(sorted(django_cachepurge.get_urls()),
                          ['/some/other/url'])

    def test_middleware(self):
        content = testingmodels.UrlModel(title=u'model with url')
        content.save()
        django_cachepurge.clear_urls()
        self.client.post('/content/1/edit', {'title': u'edited by web request'})
        self.assertEquals(self.pruned_urls,
                          ['http://127.0.0.1:3128/content/1',
                           'http://127.0.0.1:3128/some/related/url'])

    def test_middleware_with_two_caches(self):

        original_urls = middleware.CACHE_URLS
        middleware.CACHE_URLS = ('http://127.0.0.1:3128',
                                 'http://localhost:8888')
        content = testingmodels.UrlModel(title=u'model with url')
        content.save()
        django_cachepurge.clear_urls()
        self.client.post('/content/1/edit', {'title': u'edited by web request'})

        self.assertEquals(self.pruned_urls,
                          ['http://127.0.0.1:3128/content/1',
                           'http://127.0.0.1:3128/some/related/url',
                           'http://localhost:8888/content/1',
                           'http://localhost:8888/some/related/url'])
        # restore value
        middleware.CACHE_URLS = original_urls

def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(CachePurgeTestCase))
    return suite



# -*- coding: utf-8 -*-
""" This module is needed by utils
"""
from django.conf import settings

USE_HTTP_1_1_PURGE = True
try:
    USE_HTTP_1_1_PURGE = settings.USE_HTTP_1_1_PURGE
except AttributeError:
    pass



# -*- coding: utf-8 -*-
from django.conf.urls.defaults import patterns

urlpatterns = patterns(
    'django_cachepurge.tests.testingmodels',
    (r'^content/(?P<pk>\d+)/edit$', 'edit_url_model'),
    )



# -*- coding: utf-8 -*-
# From Products.CMFSquidTool.utils
# Changed: removed some Zope specific code

##############################################################################
#
# Copyright (c) 2003-2005 struktur AG and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
"""
import sys
import socket
import httplib
import urlparse
import threading
import Queue
import logging
from config import USE_HTTP_1_1_PURGE
import time

class Logger(object):

    def __init__(self, name):
        self.logger = logging.getLogger(name)

    def __getattribute__(self, name):
        if name in ('logger',):
            return object.__getattribute__(self, name)
        def wrapper(*args, **kw):
            if __name__ == '__main__' and args:
                print args[0]
            return getattr(self.logger, name)(*args, **kw)
        return wrapper

logger = Logger('cachepurge')

class Connection(httplib.HTTPConnection):

    def __init__(self, host, port=None, strict=None, scheme="http", timeout=5):
        self.scheme = scheme
        if scheme == "http":
            self.default_port = httplib.HTTP_PORT
        elif scheme == "https":
            self.default_port = httplib.HTTPS_PORT
        else:
            raise ValueError, "Invalid scheme '%s'" % (scheme,)
        httplib.HTTPConnection.__init__(self, host, port, strict)
        self.timeout = timeout

    def connect(self):
        # We want to connect in normal blocking mode, else Python on Windows
        # will timeout even for 'connection refused' style errors (fixed in
        # Python 2.4 on Windows)
        if self.scheme == "http":
            httplib.HTTPConnection.connect(self)
        elif self.scheme == "https":
            # Clone of httplib.HTTPSConnection.connect
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect((self.host, self.port))
            key_file = cert_file = None
            ssl = socket.ssl(sock, key_file, cert_file)
            self.sock = httplib.FakeSocket(sock, ssl)
        else:
            raise ValueError, "Invalid scheme '%s'" % (self.scheme,)
        # Once we have connected, set the timeout.
        self.sock.settimeout(self.timeout)

XERROR = ('x-squid-error',)

# Strategy for purging:
# Most purging is done 'asynchronously' - ie, Zope requests a purge, but
# doesn't want to block until the response is sent.  About the only use
# of a synchronous purge is explicitly via the ZMI, in which case we block
# so we can let the user see the result.
# The manual case is the exception - we don't go to huge efforts to optimize
# this case - ie, no attempt at connection-pooling, etc.
# The strategy for the async case is:
# * Each remote host gets a queue and a worker thread.
# * Each worker thread manages its own connection.  While it can not
#   establish a connection, the queue is not processed at all.  Once a
#   connection is established, the queue is purged one item at a time.
#   Should the connection fail, the worker thread again waits until a
#   connection is possible.
# This means that connections are never pooled or shared in any way (except
# as a side-effect of all requests for a single host being processed by the
# same worker-thread).
class Worker(threading.Thread):

    def __init__(self, queue, host, scheme, producer):
        self.host = host
        self.scheme = scheme
        self.producer = producer
        self.queue = queue
        self.stopping = False
        name = "PurgeThread for %s://%s" % (scheme, host)
        super(Worker, self).__init__(name=name)

    def getConnection(self, url):
        # Blocks until either a connection is established, or
        # we are asked to shut-down.  Includes a simple strategy for
        # slowing down the retry rate, retrying from 2 seconds to 2 minutes
        # until the connection appears.
        wait_time = 1
        while not self.stopping:
            try:
                return self.producer.getConnection(url)
            except socket.error, e:
                # max wait time is 2 minutes
                wait_time = min(wait_time * 2, 120)
                logger.debug("Error %s connecting to %s - will "
                             "retry in %d second(s)", e, url, wait_time)
                for i in xrange(wait_time):
                    if self.stopping:
                        break
                    time.sleep(1)
        return None # must be stopping!

    def run(self):
        logger.debug("%s starting", self)
        # Queue should always exist!
        q = self.producer.queues[(self.host, self.scheme)]
        connection = None
        try:
            while not self.stopping:
                item = q.get()
                if self.stopping or item is None: # Shut down thread signal
                    logger.debug('Stopping worker thread for '
                                 '(%s, %s).' % (self.host, self.scheme))
                    break
                url, purge_type = item

                # Loop handling errors (other than connection errors) doing
                # the actual purge.
                for i in range(5):
                    if self.stopping:
                        break
                    # Get a connection.
                    if connection is None:
                        connection = self.getConnection(url)
                        if connection is None: # stopping
                            break
                    # Got an item, prune it!
                    try:
                        resp, msg, err = self.producer._pruneUrl(connection,
                                                                 url, purge_type)
                        # worked! See if we can leave the connection open for
                        # the next item we need to process
                        # NOTE: If we make a HTTP 1.0 request to IIS, it
                        # returns a HTTP 1.1 request and closes the
                        # connection.  It is not clear if IIS is evil for
                        # not returning a "connection: close" header in this
                        # case (ie, assuming HTTP 1.0 close semantics), or
                        # if httplib.py is evil for not detecting this
                        # situation and flagging will_close.
                        if not USE_HTTP_1_1_PURGE or resp.will_close:
                            connection.close()
                            connection = None
                        break # all done with this item!

                    except (httplib.HTTPException, socket.error), e:
                        # All errors 'connection' related errors are treated
                        # the same - simply drop the connection and retry.
                        # the process for establishing the connection handles
                        # other bad things that go wrong.
                        logger.debug('Transient failure on %s for %s, '
                                     're-establishing connection and '
                                     'retrying: %s' % (purge_type, url, e))
                        connection.close()
                        connection = None
                    except:
                        # All other exceptions are evil - we just disard the
                        # item.  This prevents other logic failures etc being
                        # retried.
                        connection.close()
                        connection = None
                        logger.exception('Failed to purge %s', url)
                        break
        except:
            logger.exception('Exception in worker thread '
                             'for (%s, %s)' % (self.host, self.scheme))
        logger.debug("%s terminating", self)

class Producer(object):

    def __init__(self, factory=Connection, timeout=30, backlog=200):
        self.factory = factory
        self.timeout = timeout
        self.queues = {}
        self.workers = {}
        self.backlog = backlog
        self.queue_lock = threading.Lock()

    def getConnection(self, url):
        (scheme, host, path, params, query, fragment) = urlparse.urlparse(url)
        # Creates a new connection - result is a connection object that is
        # *already* connected - or an exception is raised by that connection
        # process.
        conn = self.factory(host, scheme=scheme, timeout=self.timeout)
        conn.connect()
        logger.debug("established connection to %s", host)
        return conn

    def getQueueAndWorker(self, url):
        (scheme, host, path, params, query, fragment) = urlparse.urlparse(url)
        key = (host, scheme)
        if key not in self.queues:
            self.queue_lock.acquire()
            try:
                if key not in self.queues:
                    logger.debug("Creating worker thread for %s://%s",
                                 scheme, host)
                    assert key not in self.workers
                    self.queues[key] = queue = Queue.Queue(self.backlog)
                    self.workers[key] = worker = Worker(queue, host, scheme, self)
                    worker.start()
            finally:
                self.queue_lock.release()
        return self.queues[key], self.workers[key]

    def stopThreads(self, wait=False):
        # Attempts to stop all threads.  Threads stop immediately after
        # the current item is being processed.
        for w in self.workers.itervalues():
            w.stopping = True
        # incase the queue is empty, wake it up so the .stopping flag is seen
        for q in self.queues.values():
            try:
                q.put(None, block=False)
            except Queue.Full:
                # no problem - self.stopping should be seen.
                pass
        ok = True
        if wait:
            for w in self.workers.itervalues():
                w.join(5)
                if w.isAlive():
                    logger.warning("Worker thread %s failed to terminate", w)
                    ok = False
        return ok

    def pruneAsync(self, url, purge_type='PURGE', daemon=True):
        (scheme, host, path, params, query, fragment) = urlparse.urlparse(url)
        __traceback_info__ = (url, purge_type, scheme, host,
                              path, params, query, fragment)

        q, w = self.getQueueAndWorker(url)
        try:
            q.put((url, purge_type), block=False)
            msg = 'Queued %s' % url
        except Queue.Full:
            # Make a loud noise.  Ideally the queue size would be
            # user-configurable - but the more likely case is that the purge
            # host is down.
            logger.warning("The purge queue for the URL %s is full - the "
                           "request will be discarded.  Please check the "
                           "server is reachable, or disable this purge host",
                           url)
            msg = "Purge queue full for %s" % url
        return msg

    def pruneUrl(self, url, purge_type='PURGE'):
        # A synchronous one targetted at the ZMI.  - just lets exceptions
        # happen, no retry semantics, etc
        logger.debug("Starting synchronous purge of %s", url)
        try:
            conn = self.getConnection(url)
            try:
                resp, xcache, xerror = self._pruneUrl(conn, url, purge_type)
                status = resp.status
            finally:
                conn.close()
        except:
            status = "ERROR"
            err, msg, tb = sys.exc_info()
            from traceback import format_exception
            xerror = '\n'.join(format_exception(err, msg, tb))
            # Avoid leaking a ref to traceback.
            del err, msg, tb
            xcache = ''
        logger.debug('Finished %s for %s: %s %s'
                     % (purge_type, url, status, xcache))
        if xerror:
            logger.debug('Error while purging %s:\n%s' % (url, xerror))
        logger.debug("Completed synchronous purge of %s", url)
        return status, xcache, xerror

    def _pruneUrl(self, conn, url, purge_type):
        (scheme, host, path, params, query, fragment) = urlparse.urlparse(url)
        __traceback_info__ = (url, purge_type, scheme, host,
                              path, params, query, fragment)
        if USE_HTTP_1_1_PURGE:
            conn._http_vsn = 11
            conn._http_vsn_str = 'HTTP/1.1'
        else:
            conn._http_vsn = 10
            conn._http_vsn_str = 'HTTP/1.0'
            # When using HTTP 1.0, to make up for the lack of a 'Host' header
            # we use the full url as the purge path, to allow for virtual
            # hosting in squid
            path = url

        purge_path = urlparse.urlunparse(('','', path, params, query, fragment))
        logger.debug('making %s request to %s for %s.' % (purge_type,
                                                          host, purge_path))
        conn.putrequest(purge_type, purge_path, skip_accept_encoding=True)
        conn.endheaders()
        resp = conn.getresponse()

        xcache = resp.getheader('x-cache', '')
        for header in XERROR:
            xerror = resp.getheader(header, '')
            if xerror:
                # Break on first found.
                break
        resp.read()
        logger.debug("%s of %s: %s %s", purge_type, url, resp.status, resp.reason)
        return resp, xcache, xerror

_producer = Producer()
pruneUrl = _producer.pruneUrl
pruneAsync = _producer.pruneAsync
stopThreads = _producer.stopThreads

if __name__ == '__main__':
    # The default logger throws things away from the cmd-line!?
    logging.basicConfig()
    logger = logging.getLogger('')
    logger.setLevel(logging.DEBUG)

    for url in sys.argv[1:]:
        pruneUrl(url)
        # Prune async
        for i in range(0, 3):
            pruneAsync(url)
    raw_input('Press enter to stop the threads...')
    stopThreads(wait=True)



# -*- coding: utf-8 -*-
""" Settings for running tests
"""
DATABASE_ENGINE = 'sqlite3'

ROOT_URLCONF = ''

CACHE_URLS = 'http://127.0.0.1:3128'

INSTALLED_APPS = (
    'django_cachepurge',
    )

MIDDLEWARE_CLASSES = (
    #'django.middleware.common.CommonMiddleware',
    'django_cachepurge.middleware.CachePurge',
    )



# -*- coding: utf-8 -*-
"""Models we want only for tests
"""
from django.db import models
from django.http import HttpResponse

class NoUrlModel(models.Model):
    """ This model has no url, it should not trigger any purge
    """
    title = models.CharField(max_length=255)

    class Meta:
        app_label = 'cachepurge_test_app'

class UrlModel(models.Model):
    """ This model has urls, it can be purged
    """
    title = models.CharField(max_length=255)

    class Meta:
        app_label = 'cachepurge_test_app'

    def get_absolute_url(self):
        return '/content/%d' % self.pk

    def get_purge_urls(self):
        return ['/some/related/url']

def edit_url_model(request, pk):

    instance = UrlModel.objects.get(pk=int(pk))
    title = request.POST['title']
    instance.title = title
    instance.save()
    return HttpResponse('edited model %s' % pk,
                        mimetype='text/plain')

class OnlyPurgeUrls(models.Model):
    """ this model doesn't have get_absolute_url, only get_purge_urls. Useful
    for join models, for example.
    """

    title = models.CharField(max_length=255)

    class Meta:
        app_label = 'cachepurge_test_app'

    def get_purge_urls(self):
        return ['/some/other/url']




import os
from setuptools import setup, find_packages

def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

version = '0.1a'

long_description = (
    read('README.txt')
    + '\n' +
    'Contributors\n'
    '============\n'
    + '\n' +
    read('CONTRIBUTORS.txt')
    + '\n' +
    'Change history\n'
    '==============\n'
    + '\n' +
    read('CHANGES.txt')
    + '\n' +
    'Download\n'
    '========\n'
    )

setup(name='django-cachepurge',
      version=version,
      description='''Django Middleware and utilities that send "PURGE" request to an upstream cache''',
      long_description=long_description,
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Environment :: Web Environment',
          'Framework :: Django',
          'Intended Audience :: Developers',
          'Topic :: Internet :: Proxy Servers',
          'License :: OSI Approved :: GNU General Public License (GPL)',
          ], # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords='django cache squid varnish',
      author='Bertrand Mathieu',
      author_email='bert.mathieu@gmail.com',
      url='http://launchpad.net/django-cachepurge',
      license='GPL',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          # -*- Extra requirements: -*-
      ],
      entry_points="""
      # -*- Entry points: -*-
      """,
      )
